arXiv is now an independent nonprofit! Learn more
License: arXiv.org perpetual non-exclusive license
arXiv:0908.4403v1 [physics.comp-ph] 30 Aug 2009

Fast calculation of HELAS amplitudes using graphics processing unit (GPU)

K. Hagiwara    J. Kanzaki thanks: e-mail: junichi.kanzaki@kek.jp    N. Okamura thanks: e-mail: naotoshi@post.kek.jp    D. Rainwater    T. Stelzer thanks: e-mail: tstelzer@uiuc.edu Affiliation: KEK Theory Center and Sokendai, Tsukuba 305-0801, Japan Affiliation: KEK, Tsukuba 305-0801, Japan Affiliation: Space and Geophysics Laboratory, Applied Research Laboratories, University of Texas, Austin, TX 78758, USA Affiliation: Dept. of Physics, University of Illinois, Urbana, IL, USA
Received: date / Revised version: August 24, 2026
Abstract

We use the graphics processing unit (GPU) for fast calculations of helicity amplitudes of physics processes. As our first attempt, we compute uu¯nγu\overline{u}\rightarrow n\gamma (n=2n\!=\!2 to 8) processes in pppp collisions at s=14\sqrt{s}=14TeV by transferring the MadGraph generated HELAS amplitudes (FORTRAN) into newly developed HEGET (HELAS Evaluation with GPU Enhanced Technology) codes written in CUDA, a C-platform developed by NVIDIA for general purpose computing on the GPU. Compared with the usual CPU programs, we obtain 40-150 times better performance on the GPU.

1 Introduction

1.1 Physics motivations

The field of particle physics is about to enter a new era of experimental discovery at the Large Hadron Collider (LHC) which will start collecting data this year. Typical new physics signals at the LHC are expected to have many high pTp_{T} jets, γ\gamma’s, WW’s and ZZ’s, and it is very important to estimate the Standard Model (SM) background to all the signals reliably. Computation of multiple particle production amplitudes is, however, often time consuming, and it is desirable to have tools that allow us to evaluate cross sections efficiently. We examine here the possibility of using a GPU (Graphic Processing Unit) to compute helicity amplitudes at many phase space points simultaneously, by taking advantage of its parallel processing capability.

1.2 GPU hardware and software

The GPU is a specialized integrated circuit (IC) used in the computer system which outputs complex images onto displays very fast. This IC is composed of many multi-processors and can process large amounts of image data in parallel with high efficiency. Because of this high performance GPUs are widely used in the scientific applications where a large number of calculations must be performed to process the data.

Recently NVIDIA [1] introduced the software development system, CUDA[2]11 1 We use CUDA version 2.1 in this paper., which enables one to develop programs which can be executed on the GPU using C/C++. We use this system to develop a new helicity amplitude calculation package which can be used on a GPU.

2 Physics process

As our first attempt we compute uu¯nγu\overline{u}\!\rightarrow\!n\gamma processes in pppp-collisions, because the amplitudes are particularly simple for numerical codes like HELAS [3] such that large numbers of Feynman diagrams, beyond the capability of automatic diagram generators like MadGraph [4], can be computed without difficulty [5].

2.1 nγn\gamma production in pppp collisions

The cross section for producing nn high pTp_{T} photons in pppp collisions can be expressed in the leading order of perturbative QCD as

dσnγ\displaystyle{\rm d\sigma}^{n\gamma} =\displaystyle= qdx1dx2[Dq/p(x1,Q)Dq¯/p(x2,Q)\displaystyle\sum_{q}\iint dx_{1}dx_{2}\left[D_{q/p}\left(x_{1},Q\right)D_{\bar{q}/p}\left(x_{2},Q\right)\right. (1)
+Dq¯/p(x1,Q)Dq/p(x2,Q)]dσ^(s^;y),\displaystyle\hskip 21.52771pt+\left.D_{\bar{q}/p}\left(x_{1},Q\right)D_{q/p}\left(x_{2},Q\right)\right]{\rm d\hat{\sigma}}\left(\hat{s};y\right)\,,

where the factorization scale QQ for the qq and q¯\overline{q} distribution functions is chosen as the pTp_{\mathrm{T}} cut-off, and the qq¯nγq\overline{q}\rightarrow n\gamma’s process cross section is

dσ^(s^)=12s^122132λ1,,λnσ1,σ2|σ1,σ2λ1,,λn|21n!dΦn.\displaystyle{\rm d\hat{\sigma}}\left(\hat{s}\right)=\dfrac{1}{2\hat{s}}\dfrac{1}{2^{2}}\dfrac{1}{3^{2}}\,3\!\!\!\!\sum_{\begin{subarray}{c}{\lambda_{1},\cdots,\lambda_{n}}\\ {\sigma_{1},\sigma_{2}}\end{subarray}}\!\!\!\!\left|{\cal{M}}^{\lambda_{1},\cdots,\lambda_{n}}_{\sigma_{1},\sigma_{2}}\right|^{2}\dfrac{1}{n!}~{\rm d\Phi_{n}}\,. (2)

Here

s^=sx1x2,\displaystyle\hat{s}=s\,x_{1}x_{2}\,, (3)

and σi\sigma_{i}(i=1,2i\!=\!1,2) denotes the helicity of the initial uu and u¯\overline{u}, λi\lambda_{i}(i=1i\!=\!1 to n) is that of the final photons. The nn-body phase space is

dΦn\displaystyle{\rm d\Phi_{n}} =\displaystyle= (2π)4δ4(p1+p2i=1nki)i=1nd3ki(2π)3 2ωi,\displaystyle\left(2\pi\right)^{4}\delta^{4}\left(p_{1}+p_{2}-\sum_{i=1}^{n}k_{i}\right)\prod_{i=1}^{n}\dfrac{d^{3}k_{i}}{(2\pi)^{3}\,2\omega_{i}}\,, (4)

where ωi=|ki|\omega_{i}=|\vec{k}_{i}| for massless photons and 1/n!1/n! is the statistical factor for nn identical photons. The helicity amplitude for the process

q(p1,σ1)+q¯(p2,σ2)γ(k1,λ1)+γ(k2,λ2)++γ(kn,λn)q(p_{1},\sigma_{1})+\overline{q}(p_{2},\sigma_{2})\rightarrow\gamma(k_{1},\lambda_{1})+\gamma(k_{2},\lambda_{2})+\cdots+\gamma(k_{n},\lambda_{n}) (5)

is particularly simple:

σ1,σ2λ1,,λn=Mσ1,σ2λ1,,λn+(n!1) permutations,\displaystyle{\cal{M}}^{\lambda_{1},\cdots,\lambda_{n}}_{\sigma_{1},\sigma_{2}}=M^{\lambda_{1},\cdots,\lambda_{n}}_{\sigma_{1},\sigma_{2}}+(n!-1)\mbox{ permutations}\,, (6)

where the full amplitude is obtained by (n!1)(n!-1) permutation of one amplitude, denoted e.g. by the Feynman diagram at Fig. 1:

Mσ1,σ2λ1,,λn\displaystyle M^{\lambda_{1},\cdots,\lambda_{n}}_{\sigma_{1},\sigma_{2}} =\displaystyle= v¯(p2,σ2)[γμn / p1 / Kn1tn1γμn1\displaystyle\bar{v}(p_{2},\sigma_{2})\left[\gamma^{\mu_{n}}\dfrac{\hbox to0.0pt{\hbox to5.03125pt{\hfil/\hfil}\hss}p_{1}-\hbox to0.0pt{\hbox to9.20833pt{\hfil/\hfil}\hss}K_{n-1}}{t_{n-1}}\gamma^{\mu_{n-1}}\cdots\right. (7)
γμ3 / p1 / K2t2γμ2 / p1 / K1t1γμ1]u(p1,σ1)\displaystyle\hskip 12.91663pt\left.\cdots\gamma^{\mu_{3}}\dfrac{\hbox to0.0pt{\hbox to5.03125pt{\hfil/\hfil}\hss}p_{1}-\hbox to0.0pt{\hbox to9.20833pt{\hfil/\hfil}\hss}K_{2}}{t_{2}}\gamma^{\mu_{2}}\dfrac{\hbox to0.0pt{\hbox to5.03125pt{\hfil/\hfil}\hss}p_{1}-\hbox to0.0pt{\hbox to9.20833pt{\hfil/\hfil}\hss}K_{1}}{t_{1}}\gamma^{\mu_{1}}\right]u(p_{1},\sigma_{1})
ϵμ1(k1,λ1)ϵμn(kn,λn)(eQq)n,\displaystyle\hskip 12.91663pt\epsilon^{\ast}_{\mu_{1}}(k_{1},\lambda_{1})\cdots\epsilon^{\ast}_{\mu_{n}}(k_{n},\lambda_{n})\,(eQ_{q})^{n}\,,

with

Kl=i=1lki,tl=(p1Kl)2.K_{l}=\sum_{i=1}^{l}k_{i},\quad t_{l}=\left(p_{1}-K_{l}\right)^{2}\,. (8)

Here QqQ_{q} is the electric charge of the quark in units of ee, the proton charge.

The standard version of MadGraph/MadEvent [4, 6, 7] generates HELAS amplitudes and computes the cross section for the nγn\gamma production process up to n=6n\!=\!6 without difficulty. The HELAS amplitude for n=7n\!=\!7 with 7!7\,! diagrams can be generated by MadGraph after enlarging the maximum size of several variables. We are unable to make MadGraph generate 8!4×1048\,!\approx 4\times 10^{4} diagrams for n=8n\!=\!8 photons, so we evaluate them by coding the amplitude of eq. (6) and by summing over (8!1)(8!-\!1) permutations of the 8 photon momenta in a numerical program.

Refer to caption
Figure 1: One of n!n! diagrams for qq¯nγq\overline{q}\to n\gamma.

2.2 Final state cuts

In order to simulate realistic LHC experiments, we introduce final state cuts for all the observed photons as follows:

pTi>pTcut=20GeV,|ηi|<ηcut=2.5,ΔRij>ΔRcut=0.4,{\begin{array}[]{rcl}p_{\mathrm{T}i}&>&p_{\mathrm{T}}^{\scriptstyle\mathrm{cut}}=20\,\mathrm{GeV},\\ |\eta_{i}|&<&\eta^{\scriptstyle\mathrm{cut}}=2.5,\\ \Delta R_{ij}&>&\Delta R^{\scriptstyle\mathrm{cut}}=0.4,\end{array}} (9)

where ηi\eta_{i} is the rapidity of the ii-th photon and

ΔRij(Δηij)2+(Δϕij)2\Delta R_{ij}\equiv\sqrt{(\Delta\eta_{ij})^{2}+(\Delta\phi_{ij})^{2}} (10)

measures rapidity and azimuthal angle separation between two photons.

For the sake of simplicity and for the purpose of the present paper, we consider only one subprocess, uu¯nγu\overline{u}\rightarrow n\gamma^{\prime}s, for the nn-photon production processes at the LHC, by neglecting contributions from dd¯d\overline{d}, ss¯s\overline{s}, cc¯c\overline{c} and bb¯b\overline{b} collisions.

3 Computation on the GPU

On the GPU, a number of programs can be executed with its multi-processors in parallel. These programs which run concurrently must be the same for all multi-processors with different input data for each program. In order to utilize the high performance of the GPU for computing the nn-photon production processes, we assign a program which processes one event on each processor.

If the program is given a set of random numbers, it generates a phase space point and computes a Jacobian and a squared amplitude for the phase space point independently from the other programs. By keeping the independence of programs among processors their structure can become simple, and we can take full advantage of the highly parallel computation on the GPU.

In this section we describe our software developed for the computation of the nn-photon productions in pppp collisions on the GPU and its computation environment.

Table 1: Parameters of GPUs
GeForce GeForce GeForce
GTX280 9800GTX 8800M GTS
Number of 30 16 8
multiprocessor
Number of core 240 128 64
Total amount of 1000 500 500
global memory [MB]
Total amount of 64 64 64
constant memory [kB]
Total amount of shared 16 16 16
memory per block [kB]
Total number of registers 16 8 8
available per block [kB]
Clock rate [GHz] 1.30 1.67 0.40

3.1 Structure of the program

For nn-photon production processes our program generates momenta and helicities of photons in the final state, those of uu and u¯\overline{u} in the initial state, and computes the total cross section of the physics process in the following order:

  1. 1.

    initialization of the program,

  2. 2.

    random number generation on the CPU,

  3. 3.

    transfer random numbers to the GPU,

  4. 4.

    generate helicities and momenta of uu, u¯\overline{u} and nγn\gamma’s using random numbers, and compute squared amplitudes on the GPU,

  5. 5.

    transfer the momenta and helicities of external particles, as well as computed squared amplitudes to the CPU, and

  6. 6.

    sum up all values to obtain the total cross section and distributions on the CPU.

Program steps between the generation of random numbers (2) and the summation of computed cross sections (6) are repeated until we obtain enough statistics for the cross section and distributions.

For the computation of physics quantities for the nn-photon production processes, we prepare two types of amplitude programs. They are:

  • a program converted from the FORTRAN program obtained with the MadGraph [4], and

  • a handwritten CUDA program using permutations of final state photons, via eq. (6).

Both programs are written using the HEGET (HELAS Evaluation with GPU Enhanced Technology) functions, newly developed codes, written in CUDA [2], to compute helicity amplitudes a la HELAS [3].

Sample codes for the case of the 3-photon production process are shown in Appendix A as List 1 and List 2 for the converted MadGraph amplitude and the permutation amplitude, respectively. They compute a sum of amplitudes for all diagrams, ampsum, a complex number, from the given momenta, {p1, \ldots, p5} and helicities, {nh1, \ldots, nh5} of the external particles. In the code for the permutation amplitude, List 2, a function, iPNext, generates a set of integers for the next permutation in the sequence. For all computation of amplitudes, new data type “cmplx”, defined in List 16, for complex numbers on GPU is used.

3.2 Execution of a CUDA function on GPU

Once a set of random numbers is generated on the CPU, computations from the generation of phase space points to the calculation of squared amplitudes are done by calling a CUDA function which is executed on the GPU. A function, which is called from a CPU program and is executed on a GPU, is called a kernel. In our program, a single call of a kernel function computes the scattering amplitudes at multiple phase-space points in parallel on a GPU.

A unit program, called a thread, which is executed on a single processor, calculates the amplitudes for one event. In the CUDA programming model a set of threads forms a thread block. Threads within one thread block can share data through the shared memory and cooperate among themselves. In the current architecture of NVIDIA’s GPUs a thread block can contain up to 512 threads. The size of a thread block can be changed within this limit when the program is executed on a GPU, and we optimize it to obtain the best performance of our program.

With a single call of a kernel, multiple thread blocks are executed in parallel on a GPU. A set of thread blocks, which is executed with a single kernel call, is called a grid. Even within a grid, threads in different thread blocks cannot share data through the shared memory.

Table 2: Configuration of the host PC
Linux PC iMac
CPU Core2Duo 3GHz Core2Duo 3.06GHz
L2 Cache 6MB 6MB
Memory 4GB 2GB
Bus Speed 1.333GHz 1.07GHz
OS Fedora 8 (64 bit) MacOS X 10.5.5
GPU GTX280 & 8800M GTS
9800GTX

3.3 Host PC environment

We tested our program on three different GPUs by NVIDIA: the GeForce GTX280, 9800GTX and 8800M GTS. The parameters for these three GPUs are given in Table 1. The GeForce 9800GTX with 128 processors was introduced in April/2008 by NVIDIA as a high-end single GPU graphic card for the high performance output of complex images like 3D to the PC display. The GeForce GTX280 with a new processor architecture, introduced in June/2008, has 240 processors, 1GByte global memory and 16k registers, which are about twice those of 9800GTX, with a 20% slower clock rate. A GeForce 8800M GTS is installed on an Apple iMac, and has 64 processors and the clock rate of 0.4GHz, but still has the same memory and register size as the 9800GTX.

Table 3: HEGET functions for external lines
External Line HEGET Function HELAS Subroutine
Flowing-In Fermion ixxxx0, ixxxx1, ixxxx2 IXXXXX
Flowing-Out Fermion oxxxx0, oxxxx1, oxxxx2 OXXXXX
Vector Boson vxxxx0, vxxxx1, vxxxx2 VXXXXX
Table 4: List of the HEGET vertex functions
Vertex Inputs Output HEGET Function HELAS Subroutine
FFV FFV Amplitude iovxxx IOVXXX
FF V jioxx0 JIOXXX
FV F fvixx0, fvoxx0 FVIXXX, FVOXXX

Parameters for the two host PCs are summarized in Table 2. They have comparable capability, although the Linux PC has twice larger memory size and 25% faster bus speed.

4 HEGET functions

In this section, we describe a set of CUDA functions which can be used on a GPU for performing helicity amplitude calculations. Based on the FORTRAN version of the HELAS library[3] we developed a set of functions, HEGET, written in CUDA.

These functions are directly converted from the HELAS subroutines into C code. We kept the order of the arguments of the HELAS subroutines but parameters for masses and widths were removed from the argument list.

The HEGET functions are organized to maximize the performance on the GPU. Since conditional branches within programs using “if” statements reduces the total efficiency of parallel computing, we eliminated them as much as possible in the HEGET functions. Accordingly, we prepared separate HEGET codes for the computation of massive and massless wave functions. In this paper we only use the HEGET functions for massless particles.

By the same token, we introduce three types of functions for external wave functions; “1” for particles moving in the +z+\!z direction, “2” for particles moving in the z-\!z direction and “0” for particles with momentum, p\vec{p}, along a generic direction. These numbers are appended to the name of the corresponding functions.

All functions relevant to this paper are listed in Table 4 and Table 4, which include all functions for massless fermions and massless vector bosons (photons), respectively. The naming scheme for HEGET functions follow that of HELAS subroutines: the HEGET (HELAS) function names start with i(I) and o(O) for flow-in and flow-out fermion wave functions, respectively, v(V) for vector boson wave functions, f(F) for off-shell fermions and j(J) for off-shell vector bosons. The correspondence between the HEGET functions and the HELAS codes are shown explicitly in Tables 4 and 4.

All of the HEGET functions that are used in this report are listed in Appendix B as List 3 to List 15. Below, we explain each function.

4.1 Fermion wave functions

We have two types of functions to compute external fermions. One is for “flowing-In” fermions and the other is for “flowing-Out” fermions.

For the flowing-In spinor wavefunction of a fermion, we prepare three functions, ixxxx1 (List 3), ixxxx2 (List 4) and ixxxx0 (List 5), derived from the HELAS subroutine, IXXXXX. These three functions compute massless fermion wavefunctions with the flowing-In fermion number. ixxxx1 is for a fermion with momentum along the +z+\!z direction, ixxxx2 is for a fermion with momentum along the z-\!z direction, and ixxxx0 is for a fermion with a generic three-momentum, p\vec{p}. All functions have the same argument list:

ixxxxk(float* p, int nHEL, int nSF, cmplx* fi) (11)

for k=1,2k=\texttt{1},\texttt{2} and 0. The inputs and outputs are:

Inputs:float p[4]4-momentumint nHELtwice fermion helicity (-1 or 1)int nSF+1 for particle, -1 for anti-particleOutputs:cmplx fi[6]fermion wavefunction |fi>u(p,nHEL/2) for nSF=+1v(p,nHEL/2) for nSF=1\begin{array}[]{ll}\textsc{Inputs:}&\\ \texttt{float p[4]}&\textrm{4-momentum}\\ \texttt{int nHEL}&\textrm{twice fermion helicity (-1 or 1)}\\ \texttt{int nSF}&\textrm{+1 for particle, -1 for anti-particle}\\ \\ \textsc{Outputs:}&\\ \texttt{cmplx fi[6]}&\textrm{fermion wavefunction {|fi>}}\\ &u(\texttt{p},\texttt{nHEL}/2)\textrm{ for }\texttt{nSF}=+1\\ &v(\texttt{p},\texttt{nHEL}/2)\textrm{ for }\texttt{nSF}=-1\\ \end{array} (12)

For the flowing-Out spinor wavefunction of a fermion, we also prepare three functions, oxxxx1 (List 6), oxxxx2 (List 7) and oxxxx0 (List 8), derived from the HELAS subroutine, OXXXXX.

These three functions compute a massless fermion wavefunction with a flowing-Out fermion number. oxxxx1 is for a fermion with momentum along the +z+\!z direction, oxxxx2 is for a fermion with momentum along z-\!z direction, and oxxxx0 is for a final state fermion with a generic three-momentum, p\vec{p}. All functions have the same argument list:

oxxxxk(float* p, int nHEL, int nSF, cmplx* fo)
(13)

for k=1,2k=\texttt{1},\texttt{2} and 0. The inputs and outputs are:

Inputs:float p[4]4-momentumint nHELtwice fermion helicity (-1 or 1)int nSF+1 for particle, -1 for anti-particleOutputs:cmplx fo[6]fermion wavefunction <fo|u¯(p,nHEL/2) for nSF=+1v¯(p,nHEL/2) for nSF=1\begin{array}[]{ll}\textsc{Inputs:}&\\ \texttt{float p[4]}&\textrm{4-momentum}\\ \texttt{int nHEL}&\textrm{twice fermion helicity (-1 or 1)}\\ \texttt{int nSF}&\textrm{+1 for particle, -1 for anti-particle}\\ \\ \textsc{Outputs:}&\\ \texttt{cmplx fo[6]}&\textrm{fermion wavefunction {<fo|}}\\ &\overline{u}(\texttt{p},\texttt{nHEL}/2)\textrm{ for }\texttt{nSF}=+1\\ &\overline{v}(\texttt{p},\texttt{nHEL}/2)\textrm{ for }\texttt{nSF}=-1\\ \end{array} (14)

4.2 Vector boson wave functions

For the wavefunction of a massless vector boson, we prepare three functions, vxxxx1 (List 9), vxxxx2 (List 10) and vxxxx0 (List 11), derived from the HELAS subroutine, VXXXXX.

These three functions compute a massless vector boson wavefunction from its four-momentum and helicity. vxxxx1 is for a vector boson with momentum along the +z+\!z direction, vxxxx2 is for a vector boson with momentum along z-\!z direction, and vxxxx0 is for a vector boson with a generic three-momentum, p\vec{p}. Their argument list is:

vxxxxk(float* p, int nHEL, int nSV, cmplx* vc) (15)

for k=1,2k=\texttt{1},\texttt{2} and 0. The inputs and outputs are:

Inputs:float p[4]4-momentumint nHELhelicity of massless vector (-1 or 1)int nSV+1 for final, -1 for initial vectorOutputs:cmplx vc[6]vector boson wavefunctionϵμ(p,nHel) for nSV=+1ϵμ(p,nHel) for nSV=1.\begin{array}[]{ll}\textsc{Inputs:}&\\ \texttt{float p[4]}&\textrm{4-momentum}\\ \texttt{int nHEL}&\textrm{helicity of massless vector (-1 or 1)}\\ \texttt{int nSV}&\textrm{+1 for final, -1 for initial vector}\\ \\ \textsc{Outputs:}&\\ \texttt{cmplx vc[6]}&\textrm{vector boson wavefunction}\\ &\epsilon^{\mu}(p,\texttt{nHel})^{*}\textrm{ for }\texttt{nSV}=+1\\ &\epsilon^{\mu}(p,\texttt{nHel})\textrm{ for }\texttt{nSV}=-1.\end{array} (16)

4.3 FFV: fermion-fermion-vector boson vertex

For the fermion-fermion-vector boson vertex,

FFV=ψ¯Fγμ[gal[0]γ52+gal[1]+γ52]ψFVμ{\cal{L}}_{\textrm{FFV}}\!=\overline{\psi}_{\textrm{F}}\gamma^{\mu}\left[\mbox{{gal[0]}}\frac{1\!-\!\gamma_{5}}{2}+\mbox{{gal[1]}}\frac{1\!+\!\gamma_{5}}{2}\right]\psi_{\textrm{F}}V_{\mu}^{*} (17)

there are four functions, iovxxx, fvixx0, fvoxx0 and jioxx0, in HEGET. They correspond to HELAS subroutines, IOVXXX, FVIXXX, FVOXXX and JIOXXX, respectively, for massless particles.

4.3.1 iovxxx

The function iovxxx (List 12) computes the amplitude of the FFV vertex from a flowing-In fermion, a flowing-Out fermion and a vector boson wave functions, whether they are on-shell or off-shell. It has the arguments:

iovxxx(cmplx* fi, cmplx* fo, cmplx* vc, float* gal, cmplx vertex)\begin{array}[]{r}\texttt{iovxxx(cmplx* fi, cmplx* fo, cmplx* vc, }\\ \texttt{float* gal, cmplx vertex)}\end{array} (18)

where the inputs and the outputs are:

Inputs:cmplx fi[6]flowing-In fermion wavefunctioncmplx fo[6]flowing-Out fermion wavefunctioncmplx vc[6]vector wavefunctionfloat gal[2]coupling constants of FFV vertexOutputs:cmplx vertexamplitude of the FFV vertex<fo|V|fi>\begin{array}[]{ll}\textsc{Inputs:}&\\ \texttt{cmplx fi[6]}&\textrm{flowing-In fermion wavefunction}\\ \texttt{cmplx fo[6]}&\textrm{flowing-Out fermion wavefunction}\\ \texttt{cmplx vc[6]}&\textrm{vector wavefunction}\\ \texttt{float gal[2]}&\textrm{coupling constants of {FFV} vertex}\\ \\ \textsc{Outputs:}&\\ \texttt{cmplx vertex}&\textrm{amplitude of the {FFV} vertex}\\ &\qquad\qquad\qquad\qquad\qquad\texttt{<fo|V|fi>}\end{array} (19)

The two chiral couplings of eq. (19) are:

gal[0]=gal[1]=eQq\texttt{gal[0]}=\texttt{gal[1]}=-eQ_{q} (20)

for the qqγqq\gamma vertex in QED, where QqQ_{q} is the electric charge of the quark in units of the proton charge, e=4παe=\sqrt{4\pi\alpha}.

4.3.2 fvixx0

The function fvixx0 (List. 13) computes the off-shell massless fermion wavefunction from a flowing-In external fermion and a vector boson. It has the arguments:

fvixx0(cmplx* fi, cmplx* vc, float* gal,cmplx* fvi)\begin{array}[]{r}\texttt{fvixx0(cmplx* fi, cmplx* vc, float* gal,}\\ \texttt{cmplx* fvi)}\end{array} (21)

where the inputs and the outputs are:

Inputs:cmplx fi[6]flowing-In fermion wavefunctioncmplx vc[6]vector wavefunctionfloat gal[2]coupling constants of the FFV vertexOutputs:cmplx fvi[6]off-shell fermion wavefunction|f’,vc,fi>\begin{array}[]{ll}\textsc{Inputs:}&\\ \texttt{cmplx fi[6]}&\textrm{flowing-In fermion wavefunction}\\ \texttt{cmplx vc[6]}&\textrm{vector wavefunction}\\ \texttt{float gal[2]}&\textrm{coupling constants of the {FFV} vertex}\\ \\ \textsc{Outputs:}&\\ \texttt{cmplx fvi[6]}&\textrm{off-shell fermion wavefunction}\\ &\texttt{|f',vc,fi>}\end{array} (22)

4.3.3 fvoxx0

The function fvoxx0 (List. 14) computes the off-shell massless fermion wavefunction from a flowing-Out external fermion and a vector boson. It has the arguments:

fvoxx0(cmplx* fo, cmplx* vc, float* gal,cmplx* fvo)\begin{array}[]{r}\texttt{fvoxx0(cmplx* fo, cmplx* vc, float* gal,}\\ \texttt{cmplx* fvo)}\end{array} (23)

where the inputs and the outputs are:

Inputs:cmplx fo[6]flowing-Out fermion wavefunctioncmplx vc[6]vector wavefunctionfloat gal[2]coupling constants of the FFV vertexOutputs:cmplx fvo[6]off-shell fermion wavefunction<fo,vc,f’|\begin{array}[]{ll}\textsc{Inputs:}&\\ \texttt{cmplx fo[6]}&\textrm{flowing-Out fermion wavefunction}\\ \texttt{cmplx vc[6]}&\textrm{vector wavefunction}\\ \texttt{float gal[2]}&\textrm{coupling constants of the {FFV} vertex}\\ \\ \textsc{Outputs:}&\\ \texttt{cmplx fvo[6]}&\textrm{off-shell fermion wavefunction}\\ &\texttt{<fo,vc,f'|}\end{array} (24)

4.3.4 jioxx0

This function, jioxx0 (List 15) computes the off-shell vector wavefunction from an external fermion pair. The off-shell vector boson wavefunction is given in the Feynman gauge for massless vectors (photons and gluons in the SM). It has the arguments:

jioxx0(cmplx* fi, cmplx* fo, float* gal,cmplx* jio)\begin{array}[]{r}\texttt{jioxx0(cmplx* fi, cmplx* fo, float* gal,}\\ \texttt{cmplx* jio)}\end{array} (25)

where the inputs and the outputs are:

Inputs:cmplx fi[6]flowing-In fermion wavefunctioncmplx fo[6]flowing-Out fermion wavefunctionfloat gal[2]coupling constants of the FFV vertexOutputs:cmplx jio[6]vector current jμ (<fo|V|fi>)\begin{array}[]{ll}\textsc{Inputs:}&\\ \texttt{cmplx fi[6]}&\textrm{flowing-In fermion wavefunction}\\ \texttt{cmplx fo[6]}&\textrm{flowing-Out fermion wavefunction}\\ \texttt{float gal[2]}&\textrm{coupling constants of the {FFV} vertex}\\ \\ \textsc{Outputs:}&\\ \texttt{cmplx jio[6]}&\textrm{vector current $j^{\mu}$ ({<fo|V|fi>})}\end{array} (26)

5 Validation of the HEGET functions

In order to validate the HEGET functions we compare the total and differential cross sections of nn-photon production processes computed on the GPU with the HEGET functions with those calculated by other programs which are based on the FORTRAN version of the HELAS library. We use MadGraph/MadEvent [4] and an independent FORTRAN program which calculates total cross section and kinematical distributions with the Monte Carlo integration program BASES [8] as references.

For comparison, we use the same physics parameters as MadGraph/MadEvent for all programs. As the parton distribution function, we use CTEQ6L1 [9] and set the factorization scale to be the pTp_{\mathrm{T}} cut of values, Q=pTcut=20GeVQ=p_{\mathrm{T}}^{\scriptstyle\mathrm{cut}}=20\,\textrm{GeV}; see eq. (9).

5.1 Total cross sections

For the calculation of the nn-photon production cross sections, the same final state cuts for all the observed photons are applied; see (eq. (9) in Sec. 2.2). Results for the computation of the total cross sections are listed in Table 5. We find the results obtained by the HEGET functions agree with those from the other programs within the statistics of generated number of events.

Up to n=5n\!=\!5, MadGraph/MadEvent gives the total cross sections and distributions, and we find agreements among all the programs. As for n=6n\!=\!6 and 7, the MadGraph generated HELAS FORTRAN codes can be integrated by using BASES [8], and the results agree well with those of HEGET codes. For n=8n\!=\!8, the FORTRAN program that performs (8!1)(8!-1) permutations of eq. (6) was used to compute the matrix elements, which were integrated by BASES.

Table 5: Total cross sections [fb] for uu¯nγu\overline{u}\to n\gamma at the LHC.
Number of photons HEGET Bases MadGraph/MadEvent
2 1.0822±0.00561.0822\pm 0.0056 1.08265±0.000311.08265\pm 0.00031 1.0811±0.00191.0811\pm 0.0019 ×104\times 10^{4}
3 6.7776±0.00826.7776\pm 0.0082 6.7849±0.00516.7849\pm 0.0051 6.775±0.0316.775\pm 0.031 ×100\times 10^{0}
4 1.2400±0.00301.2400\pm 0.0030 1.2280±0.00291.2280\pm 0.0029 1.2372±0.00411.2372\pm 0.0041 ×102\times 10^{-2}
5 2.598±0.0112.598\pm 0.011 2.596±0.0102.596\pm 0.010 2.572±0.0532.572\pm 0.053 ×105\times 10^{-5}
6 5.799±0.0175.799\pm 0.017 5.792±0.0145.792\pm 0.014 ×108\times 10^{-8}
7 1.264±0.0081.264\pm 0.008 1.261±0.0041.261\pm 0.004 ×1010\times 10^{-10}
8 2.77±0.052.77\pm 0.05 2.4±0.32.4\pm 0.3 ×1013\times 10^{-13}

5.2 Kinematical distributions

We also compare several kinematical distributions of the generated events. As an example, in Fig. 2 distributions of the maximum transverse momentum and ΔR\Delta R between final state photons for the uu¯5-photonsu\overline{u}\rightarrow 5\mbox{-photons} process are shown for the three programs. We find excellent agreement.

Refer to caption
Figure 2: Distributions of maximum transverse momentum and ΔR\Delta R of final state photons in uu¯5-photonsu\overline{u}\rightarrow 5\mbox{-photons}

Similar comparisons have been performed for all the cases listed in Table 5, and we find agreement within statistical errors.

6 Performance comparison

6.1 Register allocation and size of the thread block

Event process time is measured by the standard C library function on the CPU. We measured the time between the start of the transfer of random numbers to the GPU and the end of the transfer of computed results back to the CPU. The process time for one event is obtained by dividing the measured time by the total number of events. We also prepared a CPU program of the same structure and measured the event process time for the equivalent part of the program. The total number of events used for the measurement depends on the number of photons in the final state and the specifications of the GPU. For the GTX280, we generate 1M events for a single kernel call and repeated 100 calls of the kernel for processes with nγ5n_{\gamma}\!\leq\!5. In short, we generated 100M events for a single execution of the program.

When we compile CUDA programs, the maximum number of registers allocated to a thread can be specified. However, the total number of registers available per thread block is limited. If we allocate more registers to a thread, then the maximum number of threads in a thread block, the block size, must become smaller. We find the performance of the program depends on these two parameters, the number of registers in a thread and the number of threads in a block. Hence, we made a systematic study of the performance of these parameters.

The basic unit of the GPU processor is a multiprocessor called the Streaming Multiprocessor (SM). Threads which belong to one thread block are concurrently executed on one SM. One SM consists of eight Scalar Processors (SP) and four threads can be executed on one SP at the same time. One multiprocessor can therefore process a group of 32 threads, called a warp, in parallel.

The number of threads in a thread block of a kernel execution can be set arbitrarily up to a maximum number limited by the number of registers per thread, but the performance is found to be better when it is a multiple of the warp size, i.e. 32 threads. This is because otherwise there are idle SPs during the kernel execution.

Fig. 3 shows the dependence of the event process time for the 5-photon production process, uu¯5γu\overline{u}\rightarrow 5\gamma, on the size of thread blocks and the number of allocated registers per thread on GTX280. GTX280 has 16k registers, which are split into threads in a thread block. Three cases of numbers of registers per thread, 42, 64 and 124, are plotted in Fig. 3, which correspond to the maximum available numbers of threads in a multiple of 32 threads, 384, 256 and 128, respectively.

The dependence on the blocks size shows a clear periodicity of a multiple of 32 threads. The best performance for the 5γ5\gamma production process was observed for a combination of 64 registers per thread and 256 threads in a thread block.

Refer to caption
Figure 3: Dependence of event process time on block size (the number of threads per thread block) and the number of registers per thread on a GTX280.

6.2 Comparison of the event process time

In Fig. 4, the measured process time for one event of nn-photons production processes is shown for the GPU (GTX280) and the CPU (PC Linux with Fedora 8). They are plotted versus the number of photons in the final state.

The upper two lines show the event process time for the computation on the CPU. The programs with the MadGraph generated HELAS codes run faster than those based on the permutation amplitude for larger number of photons in the final state (nγ5n_{\gamma}\!\geq\!5). This is essentially because the MadGraph-generated codes avoid repetition of computing the same off-shell amplitudes. A larger fraction of off-shell amplitudes is repeated for larger nγn_{\gamma}, and the ratio of the process time for the MadGraph generated amplitudes and the permutation amplitudes grows from 1.7 for nγ=5n_{\gamma}\!=\!5 to 3.8 for nγ=7n_{\gamma}\!=\!7. For nγ=8n_{\gamma}\!=\!8, only the permutation-based program can be compiled on the CPU.

The lower three lines show the event process time on the GPU (GTX280). In all cases the program on the GPU computes one event much faster than the CPU. For the computation on the GPU the HEGET programs obtained by converting the HELAS codes generated by the MadGraph runs faster than those based on the permutations of a single amplitude, when nγ=3n_{\gamma}\!=\!3 or larger. The ratio grows to a factor of 2.3 for nγ=5n_{\gamma}\!=\!5.

When the number of photons becomes larger than five, the MadGraph amplitude (the HEGET code which is converted from the HELAS code generated by MadGraph) becomes very long, and the present CUDA compiler cannot process the converted program. In order to compile the long HEGET program for large nn-photon processes we divide the program into smaller pieces. Each piece computes a subset of the diagrams. The CPU program calls separate kernels for each amplitude sequentially and sums up their computed amplitudes. This method works for n=7n\!=\!7, or 7!=50407!\!=\!5040 diagrams, whose HEGET code can be compiled by dividing it into about 500 kernel calls. The 8-photon production process has so many diagrams (8!=403208!\!=\!40320), that we have not been able to compile the HEGET program even after dividing it into small pieces.

For the case of nγ=5n_{\gamma}\!\!=\!5 we examine both programs, one with a single amplitude and the other with divided amplitudes. We find that the multiple kernel program with divided amplitudes runs faster than the single kernel program for the whole amplitude by about 30%. This is probably because the size of the 5γ5\gamma amplitude program is near the maximum capability of the present CUDA compiler.

Refer to caption
Figure 4: Event process times for the GPU and CPU.

6.3 Comparison of performance of GPU and CPU

The ratios of event process times between the CPU and GPU are shown in Fig. 5. The solid curves give the ratios between the permutation amplitudes on the CPU (the upper CPU points in Fig. 4) and those on the GPU (the upper GPU points in Fig. 4). The performance ratio is about 120 for nγ=2n_{\gamma}\!=\!2 and 3, and it goes down to about 45 for nγ=6n_{\gamma}\!=\!6, 7 and 8. A higher-performance ratio is found when we compare the event process time of the MadGraph-generated amplitude on the CPU (the lower CPU points in Fig. 4) and that on the GPU (the lower GPU points in Fig. 4). Ratios above 150 are found for nγ=3n_{\gamma}\!=\!3 and 4, going down to about 45 for nγ=7n_{\gamma}\!=\!7. For nγ=6n_{\gamma}\!=\!6 and 7, the HEGET programs from the MadGraph-generated HELAS codes are too big for the present CUDA compiler, and the results are shown for the GPU program with multiple kernel calls for one event. Nevertheless, the performance ratio is about 60 for nγ=6n_{\gamma}\!=\!6 and about 45 for nγ=7n_{\gamma}\!=\!7. Even more surprisingly, the performance is better for the divided amplitude with multiple kernel calls than the single kernel computation for nγ=5n_{\gamma}\!=\!5, with the performance ratio exceeding 100.

To summarize, we find that GPU programs run faster than CPU programs by a factor exceeding 100 for nγ5n_{\gamma}\leq\!5, while a performance gain larger than 40 is achieved for other nγ8n_{\gamma}\!\leq\!8.

Refer to caption
Figure 5: Ratio of event process times (CPU/GPU).

6.4 Difference among GPUs

We also examine the performance of two other GPUs, the 9800GTX and the 8800M GTS. The GPU board with 9800GTX is connected to the same host PC as GTX280. On the other hand, the 8800M GTS is a built-in graphic card on the iMac. We ran the same programs for all GPUs and measures their event process times. Fig. 6 shows the process time per event for the permutation amplitudes on the 9800GTX and 8800M GTS relative to a GTX280, which was shown in Fig. 4 as “GPU permutation”. The ratio of process times among various GPU’s is roughly proportional to the inverse of the number of multiprocessors: 240 on the GTX280, 128 on the 9800GTX, 64 on the 8800M GTS; see Table 1. It is worth noting that even the small built-in graphic card (8800M GTS) on an iMac can compute nγn_{\gamma} production processes faster than CPU by about factors of 25, 15 and 10, respectively, for nγ=3n_{\gamma}\!=\!3, 4 and 5.

Refer to caption
Figure 6: Event process times for GPUs.

6.5 Double precision calculation

The GTX280 has one double precision processing unit for each streaming multiprocessor (SM). We prepared a double precision version of the HEGET library and the CUDA program which computes the nγn_{\gamma} cross sections with double precision. The structure of these programs is the same as the single precision version described above. Results of the single and double precision computations agree. Fig. 7 shows the ratio of event process times between double and single precision computations. Because the number of double precision processing units is small compared to the number of single precision processing units, the program with the double precision computations runs 3-4 times slower than that with single precision.

Refer to caption
Figure 7: Ratio of event process times (double/single precision) on a GTX280.

6.6 Effect of unrolling loops

In the permutation amplitude program, a combination of integer numbers for successive permutations is generated sequentially within a while loop of the C-language: see Appendix A.2 for nγ=3n_{\gamma}\!=\!3 as an example. Since programs with loops or branches lower the efficiency of parallel computing on the GPU, in general, there is a possibility of improving the performance of the permutation program by spreading out, or unrolling, a part of the while loop in the permutation.

We therefore tested the effect of unrolling the while loop for the permutation generation. Sample code for the case of the 3-photon production process is shown in Appendix C as List 17, for the amplitude program with one permutation unrolled. It should be compared with the program List 2 in Appendix A.2, which performs 3!=63!\!=\!6 permutation within one while loop. Fig. 8 shows the relative performance of the programs after unrolling a part of the while loop as compared to the process time presented in Fig. 4 as “GPU permutation”.

We find that the effect of unrolling is significant, especially for large nγn_{\gamma} processes. The improvement of the execution speed becomes a factor of 3 for nγ=7n_{\gamma}\!=\!7 and 8. Because the event process time does not change significantly by unrolling the while loop on the CPU, the CPU/GPU ratio plotted in Fig. 5 for ‘permutation‘ amplitudes grows from 45\sim 45 to 150\sim 150 for nγ=7n_{\gamma}\!=\!7 and 8 after unrolling two permutations.

Also plotted as a reference in Fig. 8 are the relative performances of the computation with the converted MadGraph amplitudes as compared to those of permutation amplitudes. It is clearly seen from Fig. 8 that by reducing the while loop the performance of the program with permutation improves significantly for nγ5n_{\gamma}\!\geq\!5, approaching that of the MadGraph based programs. The presence of the while loop is probably the main cause of the poor performance of the permutation amplitudes as compared to the MadGraph generated ones, as shown in Fig. 4.

Refer to caption
Figure 8: Effect of unrolling while loops in “permutation” amplitudes on a GTX280.

7 Summary

We have presented the results of our attempt to compute multi-particle production events at hadron colliders on GPUs [1], Graphic Processing Units. Our achievements and findings may be summarized as follows.

  • HELAS subroutines [3] written in FORTRAN were converted to HEGET functions written in CUDA [2], a C-language developed for GPU computing.

  • The HELAS amplitude code for uu¯nγu\overline{u}\rightarrow n\gamma’s (nγ7n_{\gamma}\!\leq\!7) generated by MadGraph [4] was converted to a CUDA program which calls HEGET functions.

  • CUDA programs that compute (n!1n\,!\!-\!1) permutations of a single Feynman amplitude were also made for nγ8n_{\gamma}\!\leq\!8.

  • Event process times of the GPU program on a GTX280 are more than 100 times faster than the CPU program for nγ5n_{\gamma}\!\leq\!5, while the gain is reduced to about 45 for nγ=7n_{\gamma}\!=\!7 and 8.

  • The present CUDA compiler cannot process the HEGET programs converted from the MadGraph generated HELAS codes for nγ6n_{\gamma}\!\geq\!6.

  • For nγ=5n_{\gamma}\!=\!5, 6 and 7, we were successful in running the converted MadGraph programs by dividing the whole amplitude into smaller pieces and calling them sequentially on the CPU. Performance improves by this dividing procedure for nγ=5n_{\gamma}\!=\!5, where the whole amplitude can be computed by one kernel call.

  • CUDA programs based on permutations of one HEGET amplitude are about a factor of 3 to 4 slower than those based on divided MadGraph codes for nγ5n_{\gamma}\!\geq\!5. The main cause of this slowdown was identified as the while loop in the program that generates permutation of nγn_{\gamma}-integers.

  • The same programs run on the GTX280, 9800GTX and 8800M GTS, and the performance is roughly proportional to the number of processors: 240, 128 and 64, respectively.

  • Double precision amplitudes on a GTX280 were examined, finding a factor of 2.5 to 4 slower performance compared to the single precision computation.

Acknowledgements.
Acknowledgement. We thank Johan Alwall, Qiang Li and Fabio Maltoni for stimulating discussions. This work is supported by the Grant-in-Aid for Scientific Research from the Japan Society for the Promotion of Science (No. 20340064) and the National Science Foundation (No. 0757889).

Appendix Appendix A Sample codes for amplitude calculations

Appendix A.1 Three photon production amplitude with the converted MadGraph code

List 1: uux3a(madgraph).cu
cmplx w01[6], w02[6], w03[6], w04[6], w05[6];
ixxxx1(p1, nh1, +1, w01);
oxxxx2(p2, nh2, -1, w02);
vxxxx0(p3, nh3, +1, w03);
vxxxx0(p4, nh4, +1, w04);
vxxxx0(p5, nh5, +1, w05);
cmplx w06[6], w07[6], w08[6];
cmplx ampsum = mkcmplx(0.0f, 0.0f);
cmplx amp;
fvoxx0(w02,w03,gau,w06);
fvoxx0(w06,w04,gau,w07);
iovxxx(w01,w07,w05,gau,amp);
ampsum += amp;
fvixx0(w01,w04,gau,w07);
fvoxx0(w02,w05,gau,w08);
iovxxx(w07,w08,w03,gau,amp);
ampsum += amp;
fvoxx0(w02,w03,gau,w06);
fvixx0(w01,w04,gau,w07);
iovxxx(w07,w06,w05,gau,amp);
ampsum += amp;
fvoxx0(w02,w04,gau,w06);
fvixx0(w01,w05,gau,w07);
iovxxx(w07,w06,w03,gau,amp);
ampsum += amp;
fvixx0(w01,w03,gau,w07);
fvixx0(w07,w04,gau,w08);
iovxxx(w08,w02,w05,gau,amp);
ampsum += amp;
fvixx0(w01,w03,gau,w07);
fvoxx0(w02,w04,gau,w06);
iovxxx(w07,w06,w05,gau,amp);
ampsum += amp;

Appendix A.2 Three photon production amplitude with permutations

List 2: uux3a(permutation).cu
cmplx w[5][6];
ixxxx1(p1, nh1, +1, w[0]);
oxxxx2(p2, nh2, -1, w[1]);
vxxxx0(p3, nh3, +1, w[2]);
vxxxx0(p4, nh4, +1, w[3]);
vxxxx0(p5, nh5, +1, w[4]);
cmplx w01[6],w02[6];
cmplx ampsum = mkcmplx(0.0f, 0.0f);
int Flag_Perm=1;
int a[3];
a[0] = 2;
a[1] = 3;
a[2] = 4;
do {
cmplx amp;
fvoxx0(w[1], w[a[0]], gau, w01);
fvoxx0(w01 , w[a[1]], gau, w02);
iovxx0(w[0], w02, w[a[2]], gau, amp);
ampsum += amp;
iPNext(a, 3, Flag_Perm);
} while (Flag_Perm>0);

Appendix Appendix B HEGET functions

Appendix B.1 Fermion wave functions

List 3: ixxxx1.cu
#include "cmplx.h"
__device__
void ixxxx1(float* p, int nHEL, int nSF, cmplx* fi)
{
float SQP0P3 = sqrtf(p[0]+p[3])*(float)(nSF);
int NH = nHEL*nSF;
fi[4] = mkcmplx(p[0]*(float)(nSF),
p[3]*(float)(nSF));
fi[5] = mkcmplx(p[1]*(float)(nSF),
p[2]*(float)(nSF));
cmplx CHI = mkcmplx(NH*p[1]*(1.0f/SQP0P3),
p[2]*(1.0f/SQP0P3));
cmplx CZERO = mkcmplx(0.0f, 0.0f);
cmplx CSQP0P3 = mkcmplx(SQP0P3, 0.0f);
fi[0]=(NH== 1)*CZERO + (NH==-1)*CHI;
fi[1]=(NH== 1)*CZERO + (NH==-1)*CSQP0P3;
fi[2]=(NH== 1)*CSQP0P3 + (NH==-1)*CZERO;
fi[3]=(NH== 1)*CHI + (NH==-1)*CZERO;
return;
}
List 4: ixxxx2.cu
#include "cmplx.h"
__device__
void ixxxx2(float* p, int nHEL, int nSF, cmplx* fi)
{
int NH=nHEL*nSF;
fi[4] = mkcmplx(p[0]*(float)(nSF),
p[3]*(float)(nSF));
fi[5] = mkcmplx(p[1]*(float)(nSF),
p[2]*(float)(nSF));
cmplx CHI = mkcmplx( nHEL*sqrtf(2.0f*p[0]),
0.0f);
cmplx CZERO = mkcmplx(0.0f,0.0f);
fi[0]=(NH== 1)*CZERO + (NH==-1)*CHI;
fi[1]=CZERO;
fi[2]=CZERO;
fi[3]=(NH== 1)*CHI + (NH==-1)*CZERO;
return;
}
List 5: ixxxx0.cu
#include "cmplx.h"
__device__
void ixxxx0(float* p, int nHEL, int nSF, cmplx* fi)
{
float SQP0P3 = sqrtf(p[0]+p[3])*(float)(nSF);
int NH = nHEL*nSF;
fi[4] = mkcmplx(p[0]*(float)(nSF),
p[3]*(float)(nSF));
fi[5] = mkcmplx(p[1]*(float)(nSF),
p[2]*(float)(nSF));
cmplx CHI = mkcmplx( NH*p[1]*(1.0f/SQP0P3),
(1.0f/SQP0P3)*p[2]);
cmplx CZERO = mkcmplx(0.0f,0.0f);
cmplx CSQP0P3 = mkcmplx(SQP0P3, 0.0f);
fi[0]=(NH== 1)*CZERO + (NH==-1)*CHI;
fi[1]=(NH== 1)*CZERO + (NH==-1)*CSQP0P3;
fi[2]=(NH== 1)*CSQP0P3 + (NH==-1)*CZERO;
fi[3]=(NH== 1)*CHI + (NH==-1)*CZERO;
return;
}
List 6: oxxxx1.cu
#include "cmplx.h"
__device__
void oxxxx1(float* p, int nHEL, int nSF, cmplx* fo)
{
float SQP0P3=sqrtf(p[0]+p[3])*(float)(nSF);
int NH=nHEL*nSF;
fo[4] = mkcmplx(p[0]*(float)(nSF),
p[3]*(float)(nSF));
fo[5] = mkcmplx(p[1]*(float)(nSF),
p[2]*(float)(nSF));
cmplx CHI = mkcmplx( NH*p[1]*(1.0f/SQP0P3),
-p[2]*(1.0f/SQP0P3));
cmplx CZERO = mkcmplx(0.0f, 0.0f);
cmplx CSQP0P3 = mkcmplx(SQP0P3, 0.0f);
fo[0]=(NH== 1)*CSQP0P3 + (NH==-1)*CZERO;
fo[1]=(NH== 1)*CHI + (NH==-1)*CZERO;
fo[2]=(NH== 1)*CZERO + (NH==-1)*CHI;
fo[3]=(NH== 1)*CZERO + (NH==-1)*CSQP0P3;
return;
}
List 7: oxxxx2.cu
#include "cmplx.h"
__device__
void oxxxx2(float* p, int nHEL, int nSF, cmplx* fo)
{
int NH=nHEL*nSF;
fo[4] = mkcmplx(p[0]*(float)(nSF),
p[3]*(float)(nSF));
fo[5] = mkcmplx(p[1]*(float)(nSF),
p[2]*(float)(nSF));
cmplx CHI = mkcmplx(-nHEL*sqrtf(2.0f*p[0]),
0.0f);
cmplx CZERO = mkcmplx(0.0f,0.0f);
fo[0]=CZERO;
fo[1]=(NH== 1)*CHI + (NH==-1)*CZERO;
fo[2]=(NH== 1)*CZERO + (NH==-1)*CHI;
fo[3]=CZERO;
return;
}
List 8: oxxxx0.cu
#include "cmplx.h"
__device__
void oxxxx0(float* p, int nHEL, int nSF, cmplx* fo)
{
float SQP0P3 = sqrtf(p[0]+p[3])*(float)(nSF);
int NH = nHEL*nSF;
fo[4] = mkcmplx(p[0]*(float)(nSF),
p[3]*(float)(nSF));
fo[5] = mkcmplx(p[1]*(float)(nSF),
p[2]*(float)(nSF));
cmplx CHI = mkcmplx( NH*p[1]*(1.0f/SQP0P3),
-p[2]*(1.0f/SQP0P3));
cmplx CZERO = mkcmplx(0.0f, 0.0f);
cmplx CSQP0P3 = mkcmplx(SQP0P3, 0.0f);
fo[0]=(NH== 1)*CSQP0P3 + (NH==-1)*CZERO;
fo[1]=(NH== 1)*CHI + (NH==-1)*CZERO;
fo[2]=(NH== 1)*CZERO + (NH==-1)*CHI;
fo[3]=(NH== 1)*CZERO + (NH==-1)*CSQP0P3;
return;
}

Appendix B.2 Vector boson wave functions

List 9: vxxxx1.cu
#include "cmplx.h"
__device__
void vxxxx1(float* p, int nHEL, int nSV, cmplx* vc)
{
vc[4] = mkcmplx(-p[0], -p[3]);
vc[5] = mkcmplx(-p[1], -p[2]);
vc[0] = mkcmplx(0.0f, 0.0f);
vc[3] = mkcmplx(0.0f, 0.0f);
vc[1] = mkcmplx(-(float)(nHEL)*rSQRT2, 0.0f);
vc[2] = mkcmplx( 0.0f, -rSQRT2);
return;
}
List 10: vxxxx2.cu
#include "cmplx.h"
__device__
void vxxxx2(float* p, int nHEL, int nSV, cmplx* vc)
{
vc[4] = mkcmplx(-p[0], -p[3]);
vc[5] = mkcmplx(-p[1], -p[2]);
vc[0] = mkcmplx(0.0f, 0.0f);
vc[3] = mkcmplx(0.0f, 0.0f);
vc[1] = mkcmplx(-(float)(nHEL)*rSQRT2, 0.0f);
vc[2] = mkcmplx( 0.0f, rSQRT2);
return;
}
List 11: vxxxx0.cu
#include "cmplx.h"
__device__
void vxxxx0(float* p, int nHEL, int nSV, cmplx* vc)
{
vc[4] = mkcmplx(p[0], p[3])*nSV;
vc[5] = mkcmplx(p[1], p[2])*nSV;
float rpt = rsqrtf(p[1]*p[1] + p[2]*p[2]);
vc[0] = mkcmplx(0.0f, 0.0f);
vc[3] = mkcmplx(
(float)(nHEL)*(1.0f/(rpt*p[0]))*rSQRT2,
0.0f);
float pzpt = (p[3]*(1.0f/p[0])*rpt)*rSQRT2
*(float)(nHEL);
vc[1] = mkcmplx(-p[1]*pzpt,
-nSV*p[2] * rpt * rSQRT2);
vc[2] = mkcmplx(-p[2]*pzpt,
+nSV*p[1] * rpt * rSQRT2);
return;
}

Appendix B.3 Fermion-fermion-vector boson vertex

List 12: iovxxx.cu
#include "cmplx.h"
__device__
void iovxx0(cmplx* fi, cmplx* fo, cmplx* vc,
float* gal, cmplx& vertex)
{
vertex =
gal[0]*((fo[2]*fi[0]+fo[3]*fi[1])*vc[0]
+(fo[2]*fi[1]+fo[3]*fi[0])*vc[1]
-((fo[2]*fi[1]-fo[3]*fi[0])*vc[2])
*mkcmplx(0.0f, 1.0f)
+(fo[2]*fi[0]-fo[3]*fi[1])*vc[3])
+gal[1]*((fo[0]*fi[2]+fo[1]*fi[3])*vc[0]
-(fo[0]*fi[3]+fo[1]*fi[2])*vc[1]
+((fo[0]*fi[3]-fo[1]*fi[2])*vc[2])
*mkcmplx(0.0f, 1.0f)
-(fo[0]*fi[2]-fo[1]*fi[3])*vc[3]);
return;
}
List 13: fvixx0.cu
#include "cmplx.h"
__device__
void fvixx0(cmplx* fi, cmplx* vc, float* gal,
cmplx* fvi)
{
fvi[4] = fi[4]-vc[4];
fvi[5] = fi[5]-vc[5];
float pf[4];
pf[0] = fvi[4].re;
pf[1] = fvi[5].re;
pf[2] = fvi[5].im;
pf[3] = fvi[4].im;
float pf2 = pf[0]*pf[0] -
(pf[1]*pf[1] + pf[2]*pf[2] + pf[3]*pf[3]);
cmplx cI = mkcmplx( 0.0f, 1.0f);
float d = -1.0f/pf2;
cmplx sl1 = (vc[0] + vc[3])*fi[0]
+ (vc[1] - cI*vc[2])*fi[1];
cmplx sl2 = (vc[0] - vc[3])*fi[1]
+ (vc[1] + cI*vc[2])*fi[0];
cmplx sr1 = (vc[0] - vc[3])*fi[2]
- (vc[1] - cI*vc[2])*fi[3];
cmplx sr2 = (vc[0] + vc[3])*fi[3]
- (vc[1] + cI*vc[2])*fi[2];
fvi[0] = ( gal[0]*((pf[0]-pf[3])*sl1
- conj(fvi[5])*sl2))*d;
fvi[1] = ( gal[0]*((pf[0]+pf[3])*sl2
- fvi[5] *sl1))*d;
fvi[2] = ( gal[1]*((pf[0]+pf[3])*sr1
+ conj(fvi[5])*sr2))*d;
fvi[3] = ( gal[1]*((pf[0]-pf[3])*sr2
+ fvi[5] *sr1))*d;
return;
}
List 14: fvoxx0.cu
#include "cmplx.h"
__device__
void fvoxx0(cmplx* fo, cmplx* vc, float* gal,
cmplx* fvo)
{
fvo[4] = fo[4]+vc[4];
fvo[5] = fo[5]+vc[5];
float pf[4];
pf[0] = fvo[4].re;
pf[1] = fvo[5].re;
pf[2] = fvo[5].im;
pf[3] = fvo[4].im;
float pf2 = pf[0]*pf[0]
- (pf[1]*pf[1] + pf[2]*pf[2] + pf[3]*pf[3]);
cmplx cI = mkcmplx( 0.0f, 1.0f);
float d = -1.0f/pf2;
cmplx sl1 = (vc[0] + vc[3])*fo[2]
+ (vc[1] + cI*vc[2])*fo[3];
cmplx sl2 = (vc[0] - vc[3])*fo[3]
+ (vc[1] - cI*vc[2])*fo[2];
cmplx sr1 = (vc[0] - vc[3])*fo[0]
- (vc[1] + cI*vc[2])*fo[1];
cmplx sr2 = (vc[0] + vc[3])*fo[1]
- (vc[1] - cI*vc[2])*fo[0];
fvo[0] = ( gal[1]*((pf[0]+pf[3])*sr1
+ fvo[5] *sr2 ))*d;
fvo[1] = ( gal[1]*((pf[0]-pf[3])*sr2
+ conj(fvo[5])*sr1 ))*d;
fvo[2] = ( gal[0]*((pf[0]-pf[3])*sl1
- fvo[5] *sl2 ))*d;
fvo[3] = ( gal[0]*((pf[0]+pf[3])*sl2
- conj(fvo[5])*sl1 ))*d;
return;
}
List 15: jioxx0.cu
#include "cmplx.h"
__device__
void jioxx0(cmplx* fi, cmplx* fo, float* gal,
cmplx* jio)
{
jio[4]=fo[4]-fi[4];
jio[5]=fo[5]-fi[5];
float DD=1.0f/(
(jio[4].re)*(jio[4].re) - (jio[5].re)*(jio[5].re)
- (jio[5].im)*(jio[5].im) - (jio[4].im)*(jio[4].im));
jio[0]=(gal[0]*(fo[2]*fi[0]+fo[3]*fi[1])
+gal[1]*(fo[0]*fi[2]+fo[1]*fi[3]))*DD;
jio[1]=(gal[1]*(fo[0]*fi[3]+fo[1]*fi[2])
-gal[0]*(fo[2]*fi[1]+fo[3]*fi[0]))*DD;
jio[2]=(gal[0]*(fo[2]*fi[1]-fo[3]*fi[0])
-gal[1]*(fo[0]*fi[3]-fo[1]*fi[2]))
*mkcmplx(0.0f,DD);
jio[3]=(gal[1]*(fo[0]*fi[2]-fo[1]*fi[3])
-gal[0]*(fo[2]*fi[0]-fo[3]*fi[1]))*DD;
return;
}

Appendix B.4 Complex numbers

List 16: cmplx.h
#ifndef __cmplx_h__
#define __cmplx_h__
typedef struct __align__(8){
float re;
float im;
} cmplx;
inline __host__ __device__
cmplx mkcmplx(float re, float im){
cmplx z;
z.re=re;
z.im=im;
return z;
}
inline __host__ __device__
cmplx mkcmplx(cmplx z){
return mkcmplx(z.re,z.im);
}
inline __host__ __device__
cmplx mkcmplx(float a){
return mkcmplx(a,0.0);
}
inline __host__ __device__
float real(cmplx a){
return a.re;
}
inline __host__ __device__
float imag(cmplx a){
return a.im;
}
inline __host__ __device__
cmplx conj(cmplx a){
return mkcmplx(a.re,-a.im);
}
inline __host__ __device__
cmplx operator+(cmplx a, cmplx b){
return mkcmplx(a.re + b.re, a.im + b.im);
}
inline __host__ __device__
void operator+=(cmplx &a, cmplx b){
a.re += b.re; a.im += b.im;
}
inline __host__ __device__
cmplx operator+(cmplx a){
return mkcmplx(+a.re, +a.im);
}
inline __host__ __device__
cmplx operator-(cmplx a, cmplx b){
return mkcmplx(a.re - b.re, a.im - b.im);
}
inline __host__ __device__
void operator-=(cmplx &a, cmplx b){
a.re -= b.re; a.im -= b.im;
}
inline __host__ __device__
cmplx operator-(cmplx a){
return mkcmplx(-a.re, -a.im);
}
inline __host__ __device__
cmplx operator*(cmplx a, cmplx b){
return mkcmplx((a.re * b.re) - (a.im * b.im),
(a.re * b.im) + (a.im * b.re));
}
inline __host__ __device__
cmplx operator*(cmplx a, float s){
return mkcmplx(a.re * s, a.im * s);
}
inline __host__ __device__
cmplx operator*(float s, cmplx a){
return mkcmplx(a.re * s, a.im * s);
}
inline __host__ __device__
void operator*=(cmplx &a, float s){
a.re *= s; a.im *= s;
}
inline __host__ __device__
cmplx operator/(cmplx a, cmplx b){
float tmpD=(1.0f/(b.re*b.re+b.im*b.im));
return mkcmplx(
( (a.re * b.re) + (a.im * b.im))*tmpD,
(-(a.re * b.im) + (a.im * b.re))*tmpD
);
}
inline __host__ __device__
cmplx operator/(cmplx a, float s){
float inv = 1.0f / s;
return a * inv;
}
inline __host__ __device__
cmplx operator/(float s, cmplx a){
float inv = s*(1.0f/(a.re*a.re+a.im*a.im));
return mkcmplx(inv*a.re,-inv*a.im);
}
inline __host__ __device__
void operator/=(cmplx &a, float s){
float inv = 1.0f / s;
a *= inv;
}
inline __host__ __device__
float fabsc(cmplx a){
return sqrtf((a.re*a.re)+(a.im*a.im));
}
inline __host__ __device__
float fabs2c(cmplx a){
return (a.re*a.re)+(a.im*a.im);
}
#endif

Appendix Appendix C A sample code for unrolling the permutation loop for three photon production process

List 17: uux3a(unrolled).cu
cmplx w[5][6];
ixxxx1(p1, nh1, +1, w[0]);
oxxxx2(p2, nh2, -1, w[1]);
vxxxx0(p3, nh3, +1, w[2]);
vxxxx0(p4, nh4, +1, w[3]);
vxxxx0(p5, nh5, +1, w[4]);
cmplx w01[6],w02[6];
cmplx ampsum = mkcmplx(0.0f, 0.0f);
int Flag_Perm=1;
int a[3];
a[0] = 2;
a[1] = 3;
a[2] = 4;
do {
cmplx amp;
fvoxx0(w[1], w[a[0]], gau, w09);
fvoxx0(w09 , w[a[1]], gau, w10);
iovxx0(w[0], w10, w[a[2]], gau, amp);
ampsum += amp;
fvoxx0(w[1], w[a[0]], gau, w09);
fvoxx0(w09 , w[a[2]], gau, w10);
iovxx0(w[0], w10, w[a[1]], gau, amp);
ampsum += amp;
fvoxx0(w[1], w[a[2]], gau, w09);
fvoxx0(w09 , w[a[0]], gau, w10);
iovxx0(w[0], w10, w[a[1]], gau, amp);
ampsum += amp;
iPNext(a, 2, Flag_Perm);
} while (Flag_Perm>0);

References

  • (1) http://www.nvidia.com/page/home.html
  • (2) http://www.nvidia.com/object/cuda_home.html We use the CUDA version 2.1 in this paper.
  • (3) K. Hagiwara, H. Murayama and I. Watanabe, Nucl. Phys. B367 (1991) 257; H. Murayama, I. Watanabe and K. Hagiwara, KEK-Report 91-11, 1992.
  • (4) T. Stelzer and W. F. Long, Comput. Phys. Commun. 81 (1994) 357.
  • (5) R. Kleiss and W. J. Stirling, Phys. Lett. B179 (1986) 159; N. Brown, K. Hagiwara and A. D. Martin, Nucl. Phys. B288 (1987) 782.
  • (6) F. Maltoni and T. Stelzer, JHEP 0302 (2003) 027.
  • (7) J. Alwall et al., JHEP 0709 (2007) 028.
  • (8) S. Kawabata, Comput. Phys. Commun. 41 (1986) 127.
  • (9) CTEQ Collaboration, H.L. Lai et al., Eur. Phys. J. C12 (2000) 375.